@@ -1,7 +1,7 @@
 
 /*  $Header$
  *
- *  Copyright (c) 1987-2007 Sun Microsystems, Inc. All Rights Reserved.
+ *  Copyright (c) 1987-2008 Sun Microsystems, Inc. All Rights Reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -194,6 +194,15 @@ perform_undo(void)
     refresh_display(e->cursor);
 }
 
+struct exprm_state *
+peek_previous_state(void)
+{
+    if (v->h.current != v->h.begin) {
+        return &(v->h.e[(v->h.current - 1) % UNDO_HISTORY_LENGTH]);
+    } else {
+        return NULL;
+    }
+}
 
 static int
 is_undo_step()
@@ -273,7 +282,7 @@ exp_insert(char *text, int cursor)
 
 
 void
-exp_del()
+exp_clear()
 {
     exp_replace("Ans");
 }
@@ -319,7 +328,8 @@ exp_backspace(int cursor)
     char buf[MAXLINE] = "", *display;
     struct exprm_state *e = get_state();
     int i, MP_reg[MP_SIZE];
-    
+
+    /* If cursor is at end of the line then delete the last character preserving accuracy */
     if (cursor < 0) {
         if (exp_has_postfix(e->expression, "Ans")) {
             char *ans = make_number(e->ans, v->base, FALSE);
@@ -351,6 +361,21 @@ exp_backspace(int cursor)
 }
 
 
+static int
+exp_delete(int cursor)
+{
+    char buf[MAXLINE] = "", *display;
+    
+    if (cursor >= 0) {
+        display = ui_get_display();
+        SNPRINTF(buf, MAXLINE, "%.*s%s", cursor, display, display + cursor + 1);
+        exp_replace(buf);
+    }
+
+    return cursor;
+}
+
+
 void
 exp_replace(char *text)
 {
@@ -444,12 +469,14 @@ void
 do_expression(int function, int arg, int cursor)
 {
     char buf[MAXLINE];
-    struct exprm_state *e;
+    struct exprm_state *e, *ex;
     
     new_state();
     e = get_state();
+    ex = peek_previous_state();
 
     e->cursor = cursor;
+    e->clear = 0;
 
     ui_set_statusbar("", "");
 
@@ -463,7 +490,8 @@ do_expression(int function, int arg, int cursor)
     switch (buttons[function].id) {
         case KEY_CLEAR:
         case KEY_CLEAR_ENTRY:
-            exp_del();
+            e->clear = 1;
+            exp_clear();
             ui_set_error_state(FALSE);
             MPstr_to_num("0", DEC, e->ans);
             break;
@@ -501,6 +529,10 @@ do_expression(int function, int arg, int cursor)
         case KEY_BACKSPACE:
             cursor = exp_backspace(cursor);
             break;
+        
+        case KEY_DELETE:
+            cursor = exp_delete(cursor);
+            break;
 
         case KEY_CHANGE_SIGN:
             exp_negate();
@@ -577,15 +609,25 @@ do_expression(int function, int arg, int cursor)
         default:
             /* If display is a number then perform functions on that number */
             if ((buttons[function].flags & (PREFIXOP | FUNC))
-                && strcmp(e->expression, "Ans") == 0) {
+                && !strcmp(e->expression, "Ans")
+                && !(ex && ex->clear)) {
                 SNPRINTF(buf, MAXLINE, "%s(%s)",
                          buttons[function].symname,
                          e->expression);
                 exp_replace(buf);
+            } else if ((((buttons[function].id == KEY_ADD) 
+                         || (buttons[function].id == KEY_SUBTRACT))
+                        && !strcmp(e->expression, "Ans")
+                        && (ex && ex->clear))) {
+                exp_replace(buttons[function].symname);
             } else {
                 if (buttons[function].flags & FUNC) {
                     SNPRINTF(buf, MAXLINE, "%s(", buttons[function].symname);
-                    cursor = exp_insert(buf, cursor);
+                    if (!ex || ex->clear) {
+                        exp_replace(buf);
+                    } else {
+                        cursor = exp_insert(buf, cursor);
+                    }
                 } else {
                     cursor = exp_insert(buttons[function].symname, cursor);
                 }
@@ -650,16 +692,24 @@ do_calc()      /* Perform arithmetic calculation and display result. */
             break;
 
         case KEY_MODULUS_DIVIDE:
-            mpcmim(v->MPresult, MP1);
-            mpcmim(v->MPdisp_val, MP2);
-            if (!mpeq(v->MPresult, MP1) || !mpeq(v->MPdisp_val, MP2)) {
-                doerr(_("Error, operands must be integers"));
-            }
+            if (!is_integer(v->MPresult) || !is_integer(v->MPdisp_val)) {
+                ui_set_statusbar(_("Error, operands must be integers"),
+                                 "gtk-dialog-error");
+            } else {
+                int val;
 
-            mpdiv(v->MPresult, v->MPdisp_val, MP1);
-            mpcmim(MP1, MP1);
-            mpmul(MP1, v->MPdisp_val, MP2);
-            mpsub(v->MPresult, MP2, v->MPresult);
+                mpdiv(v->MPresult, v->MPdisp_val, MP1);
+                mpcmim(MP1, MP1);
+                mpmul(MP1, v->MPdisp_val, MP2);
+                mpsub(v->MPresult, MP2, v->MPresult);
+
+                val = 0;
+                mpcim(&val, MP1);
+                if ((mplt(v->MPdisp_val, MP1) && mpgt(v->MPresult, MP1)) ||
+                    mplt(v->MPresult, MP1)) { 
+                    mpadd(v->MPresult, v->MPdisp_val, v->MPresult);
+                }
+            }
             break;
 
         case KEY_X_POW_Y:
@@ -909,7 +959,7 @@ do_constant(int index)
 
 /* Remove the last numeric character typed. */
 void
-do_delete()
+do_backspace()
 {
     size_t len;
 
@@ -943,6 +993,13 @@ do_delete()
 }
 
 
+void
+do_delete()
+{
+    /* Not required in ltr mode */
+}
+
+
 /* Exchange display with memory register. */
 void
 do_exchange(int index)
@@ -1119,7 +1176,11 @@ do_immedfunc(int s[MP_SIZE], int t[MP_SIZE])
             break;
 
         case KEY_LOGARITHM:
-            mplog10(s, t);
+            mplogn(10, s, t);
+            break;
+
+        case KEY_LOGARITHM2:
+            mplogn(2, s, t);
             break;
 
         case KEY_RANDOM:
